/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <ctype.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <gtk/gtk.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <fcntl.h>
#include "tcpjunk.h"

struct gui_elements
  {
    GtkWidget *window;
    GtkWidget *host_entry;
    GtkWidget *port_entry;
    GtkWidget *session_text;
    GtkWidget *session_log_text;
    GtkWidget *connect_butt;
    GtkWidget *listen_butt;
    GtkWidget *status_text;
    GtkWidget *ses_set_lab;
    GtkWidget *init_to_entry;
    GtkWidget *sock_to_entry;
    GtkWidget *repeat_entry;
    GtkWidget *syntax_entry;
    GtkWidget *open_butt;
    char srv_addr[FILENAME_MAX];
    char *cur_folder;
    int auto_scroll;
    int auto_clearstat;
    int gui_listen_sock;
    int listening;
  };

struct gui_elements gui_data;
struct tj_data gui_instance_data;

static gboolean delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
  {
  gtk_main_quit();
  return FALSE;
  }

static void gui_update()
  {

  while (gtk_events_pending())
    gtk_main_iteration();

  }

static void gui_clear_session_log()
  {
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gtk_text_buffer_delete(buffer, &start, &end);
  gui_update();
  }

static void gui_clear_status()
  {
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  gtk_text_buffer_delete(buffer, &start, &end);
  gui_update();
  }

static void gui_scroll_session_log()
  {

  /* auto scroll */
  GtkTextBuffer *buffer;
  GtkTextIter end;

  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  gtk_text_buffer_get_end_iter(buffer, &end);
  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(gui_data.session_log_text), &end, 0, TRUE, 0, 0);
  }

static void gui_show_selection(GtkWidget *text_view)
  {

  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(text_view));
  gtk_text_buffer_get_selection_bounds(buffer, &start, &end);
  char tempbuf[32];
  sprintf(tempbuf, "Copied %d characters", gtk_text_iter_get_offset(&end)
      - gtk_text_iter_get_offset(&start));
  tj_error(&gui_instance_data, tempbuf, "", 0);

  }

static void syntax_highlight()
  {
  GtkTextBuffer *ses_txt_buf, *ses_log_buf;
  GtkTextIter st_start, st_end, sl_start, sl_end;
  GtkTextTag *ses_text_tag, *ses_log_tag;
  GtkTextIter mstart, mend;
  GtkTextTagTable *tag_table;
  gchar *hl_text;
  int matches = 0;
  char match_msg[128];

  hl_text = (gchar *) gtk_entry_get_text(GTK_ENTRY(gui_data.syntax_entry));

  /* clear existing tags */
  ses_txt_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_text));
  gtk_text_buffer_get_bounds(ses_txt_buf, &st_start, &st_end);
  gtk_text_buffer_remove_tag_by_name(ses_txt_buf, "yellow_hl", &st_start,
      &st_end);
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  gtk_text_buffer_get_bounds(ses_log_buf, &sl_start, &sl_end);
  gtk_text_buffer_remove_tag_by_name(ses_log_buf, "yellow_hl", &sl_start,
      &sl_end);

  if (strlen(hl_text) < 1)
    return;

  tag_table = gtk_text_buffer_get_tag_table(ses_txt_buf);
  ses_text_tag = gtk_text_tag_table_lookup(tag_table, "yellow_hl");
  tag_table = gtk_text_buffer_get_tag_table(ses_log_buf);
  ses_log_tag = gtk_text_tag_table_lookup(tag_table, "yellow_hl");

  /* session text */
  while (gtk_text_iter_forward_search(&st_start, hl_text, 0, &mstart, &mend,
      NULL))
    {
    matches++;
    gtk_text_buffer_apply_tag(ses_txt_buf, ses_text_tag, &mstart, &mend);
    st_start = mend;
    }
  if (matches)
    {
    sprintf(match_msg, "%d highlights in session text", matches);
    tj_error(&gui_instance_data, match_msg, "", 0);
    matches = 0;
    }

  /* session log */
  while (gtk_text_iter_forward_search(&sl_start, hl_text, 0, &mstart, &mend,
      NULL))
    {
    matches++;
    gtk_text_buffer_apply_tag(ses_log_buf, ses_log_tag, &mstart, &mend);
    sl_start = mend;
    }
  if (matches)
    {
    sprintf(match_msg, "%d highlights in session log", matches);
    tj_error(&gui_instance_data, match_msg, "", 0);
    }

  gui_update();

  }

static void gui_connect(GtkWidget *connect_butt)
  {

  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), FALSE);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.listen_butt), FALSE);
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
  gui_clear_status();
  else
  tj_error(&gui_instance_data, "-", "", 0);

  /* check address */
  int check_ok = 0;
  char *temp_str;
  temp_str = (gchar *) gtk_entry_get_text(GTK_ENTRY(gui_data.host_entry));
  strcpy(gui_data.srv_addr, temp_str);
  check_ok = check_address(&gui_instance_data, gui_data.srv_addr);
  if (check_ok != 1)
  goto done;

  /* check port */
  gui_instance_data.dest_port = atoi(gtk_entry_get_text(GTK_ENTRY(
              gui_data.port_entry)));
  check_ok = check_port(gui_instance_data.dest_port, &gui_instance_data);
  if (check_ok != 1)
  goto done;

  /* timeouts */
  gui_instance_data.itimeout = atoi(gtk_entry_get_text(GTK_ENTRY(
              gui_data.init_to_entry)));
  gui_instance_data.stimeout = atoi(gtk_entry_get_text(GTK_ENTRY(
              gui_data.sock_to_entry)));
  gui_instance_data.stimeout *= 1000;

  /* repeat */
  gui_instance_data.keepalive = atoi(gtk_entry_get_text(GTK_ENTRY(
              gui_data.repeat_entry)));

  /* load session data from GUI */
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_text));
  gtk_text_buffer_get_bounds(buffer, &start, &end);
  char *gui_session_text =
  gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

  /* check session data */
  gui_instance_data.session_data_len = strlen(gui_session_text);
  if(!gui_instance_data.session_data_len)
    {
    tj_error(&gui_instance_data, "Session data is empty", "", 1);
    goto done;
    }
  gui_instance_data.session_data = tj_realloc(gui_instance_data.session_data,
      gui_instance_data.session_data_len);
  memcpy(gui_instance_data.session_data, gui_session_text,
      gui_instance_data.session_data_len);
  find_sections(&gui_instance_data);

  /* connect */
  junk(gui_instance_data);
  syntax_highlight();

  /* finish */
  done:

  if (gui_data.auto_scroll == 1)
    {
    gui_scroll_session_log();
    }
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), TRUE);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.listen_butt), TRUE);
  return;

  }

static gboolean gui_check_accept(gpointer fdptr)
  {
  if (gui_data.listening == 0)
    return FALSE;

  /* initialize descriptors set */
  int sel_ret;
  fd_set set;
  FD_ZERO (&set);
  FD_SET (gui_data.gui_listen_sock, &set);

  /* initialize timeout */
  struct timeval timeout;
  timeout.tv_sec = 0;
  timeout.tv_usec = 50;

  sel_ret = select(gui_data.gui_listen_sock + 1, &set, NULL, NULL, &timeout);

  if (sel_ret < 1)
  return TRUE;
  else
    {
    gui_accept();
    return FALSE;
    }
  }

static gint syntax_entry_key_press(GtkWidget *widget, GdkEventKey *key,
    gpointer data)
  {
  guint keyval;
  keyval = key->keyval;

  switch (keyval)
    {
    case GDK_Return:
    case GDK_KP_Enter:
    case GDK_ISO_Enter:
    case GDK_3270_Enter:
      syntax_highlight();

    default:
      return 0;
    }
  }

static void gui_enable(gboolean state)
  {
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.connect_butt), state);
  gtk_widget_set_sensitive(GTK_WIDGET(gui_data.open_butt), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.host_entry), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.port_entry), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.init_to_entry), state);
  gtk_entry_set_editable(GTK_ENTRY(gui_data.sock_to_entry), state);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(gui_data.session_text), state);
  }

static void gui_stop_listening()
  {
  gui_data.listening = 0;
  gtk_button_set_label(GTK_BUTTON(gui_data.listen_butt), "Listen");
  gui_enable(TRUE);

  close(gui_data.gui_listen_sock);
  return;
  }

void gui_accept()
  {
  /* accept */
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
    gui_clear_status();
  unsigned int sin_size = sizeof(struct sockaddr_in);
  struct sockaddr_in rsa;

  gui_instance_data.new_sock = accept(gui_data.gui_listen_sock,
      (struct sockaddr *) &rsa, &sin_size);
  if (gui_instance_data.new_sock < 0)
    {
    tj_error(&gui_instance_data, "Failed accept on socket: ", strerror(errno),
        1);
    gui_stop_listening();
    return;
    }

  /* start transfers */
  tj_error(&gui_instance_data, "Connection accepted", "", 0);

  /* stop checking */
  g_timeout_add(500, (GSourceFunc) gui_check_accept, NULL);

  /* start session */
  sjunk(gui_instance_data);

  if (gui_data.auto_scroll == 1)
    {
    gui_scroll_session_log();
    }
  }

static void gui_listen(GtkWidget *listen_butt)
  {
  if (gui_data.listening == 1)
    {
    gui_stop_listening();
    return;
    }

  /* disable GUI */
  gui_enable(FALSE);
  gtk_button_set_label(GTK_BUTTON(gui_data.listen_butt), "Stop listening");

  /* clear GUI */
  gui_clear_session_log();
  if (gui_data.auto_clearstat == 1)
    gui_clear_status();
  int ret;

  /* check port */
  gui_instance_data.dest_port = atoi(gtk_entry_get_text(GTK_ENTRY(gui_data.port_entry)));
  ret = check_port(gui_instance_data.dest_port, &gui_instance_data);
  if (ret != 1)
    goto failure;

  /* timeouts */
  gui_instance_data.itimeout = atoi(gtk_entry_get_text(GTK_ENTRY(gui_data.init_to_entry)));
  gui_instance_data.stimeout = atoi(gtk_entry_get_text(GTK_ENTRY(gui_data.sock_to_entry)));
  gui_instance_data.stimeout *= 1000;

  /* keep alive */
  gui_instance_data.keepalive = atoi(gtk_entry_get_text(GTK_ENTRY(gui_data.repeat_entry)));

  /* get session data from GUI*/
  GtkTextIter start, end;
  GtkTextBuffer *buffer;
  buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_text));
  gtk_text_buffer_get_bounds(buffer, &start, &end);

  char *gui_session_text =
      gtk_text_buffer_get_text(buffer, &start, &end, FALSE);

  /* check session data */
  gui_instance_data.session_data_len = strlen(gui_session_text);
  if (!gui_instance_data.session_data_len)
    {
    tj_error(&gui_instance_data, "Session data is empty", "", 1);
    goto failure;
    }
  gui_instance_data.session_data = tj_realloc(gui_instance_data.session_data,
      gui_instance_data.session_data_len);
  memcpy(gui_instance_data.session_data, gui_session_text,
      gui_instance_data.session_data_len);
  find_sections(&gui_instance_data);

  /* preparations */
  struct sockaddr_in sa;
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = INADDR_ANY;
  sa.sin_port = htons(gui_instance_data.dest_port);
  int yes = 1;

  /* create listening socket */
  gui_data.gui_listen_sock = socket(AF_INET,SOCK_STREAM, 0);
  if (gui_data.gui_listen_sock < 0)
    {
    tj_error(&gui_instance_data, "Failed to allocate socket: ",
        strerror(errno), 1);
    goto failure;
    }

  /* set socket reuse */
  ret = setsockopt(gui_data.gui_listen_sock, SOL_SOCKET, SO_REUSEADDR, &yes,
      sizeof(int));
  if (ret < 0)
    {
    tj_error(&gui_instance_data, "Error setting socket options: ", strerror(
        errno), 1);
    goto failure;
    }

  /* bind socket */
  ret = bind(gui_data.gui_listen_sock, (struct sockaddr *) &sa,
      sizeof(struct sockaddr));
  if (ret < 0)
    {
    tj_error(&gui_instance_data, "Failed to bind socket: ", strerror(errno), 1);
    goto failure;
    }

  /* listen on socket */
  ret = listen(gui_data.gui_listen_sock, BACKLOG);
  if (ret < 0)
    {
    tj_error(&gui_instance_data, "Failed to listen on socket: ",
        strerror(errno), 1);
    goto failure;
    }

  /* all was OK */
  g_timeout_add(250, (GSourceFunc) gui_check_accept, NULL);
  gui_data.listening = 1;
  return;

  /* something wrong */
  failure:

  gui_stop_listening();
  return;

  }

void gui_show_traffic(char *buf, int buf_len, int color)
  {
  GtkTextBuffer *ses_log_buf;
  GtkTextIter start, end;
  GtkTextTag *fg_tag;
  GtkTextTagTable *tag_table;
  int i;

  /* disable unprintable chars */
  for (i = 0; i < buf_len; i++)
    {
    if (isprint(buf[i]) < 1 && isspace(buf[i]) < 1)
      buf[i] = '.';
    }

  /* get existing text */
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  gtk_text_buffer_get_bounds(ses_log_buf, &start, &end);
  tag_table = gtk_text_buffer_get_tag_table(ses_log_buf);

  /* choose a color */
  if (color == 0)
    fg_tag = gtk_text_tag_table_lookup(tag_table, "red_fg");
  else
    fg_tag = gtk_text_tag_table_lookup(tag_table, "blue_fg");

  /* insert text */
  gtk_text_buffer_insert_with_tags(ses_log_buf, &end, buf, buf_len, fg_tag,
      NULL);

  }

void gui_add_log(char *buf, int color)
  {
  GtkTextBuffer *stat_buf;
  GtkTextIter end;
  GtkTextTag *fg_tag;
  GtkTextTagTable *tag_table;

  stat_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_get_end_iter(stat_buf, &end);
  tag_table = gtk_text_buffer_get_tag_table(stat_buf);

  /* choose a color */
  if (color == 0)
    gtk_text_buffer_insert(stat_buf, &end, buf, -1);
  else
    {
    fg_tag = gtk_text_tag_table_lookup(tag_table, "red_fg");
    gtk_text_buffer_insert_with_tags(stat_buf, &end, buf, -1, fg_tag, NULL);
    }

  gtk_text_view_scroll_to_iter(GTK_TEXT_VIEW(gui_data.status_text), &end, 0, TRUE, 0, 0);

  gui_update();
  }

void gui_file_open(GtkTextView *session_text)
  {

  /* open file dialog */
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new("Open File", GTK_WINDOW(gui_data.window),
  GTK_FILE_CHOOSER_ACTION_OPEN, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
      GTK_STOCK_OPEN, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), gui_data.cur_folder);
  if (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
    char *file_name;
    file_name = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));

    /* load config file */
    strcpy(gui_instance_data.session_file_name, file_name);
    load_session_data(&gui_instance_data);

    /* disable unprintable chars */
    int i;
    int binary_found = 0;
    for (i = 0; i < gui_instance_data.session_data_len; i++)
      {
      if (isprint(gui_instance_data.session_data[i]) < 1 && isspace(gui_instance_data.session_data[i]) < 1)
        {
        gui_instance_data.session_data[i] = '.';
        binary_found = 1;
        }
      }
    if (binary_found)
      tj_error(
          &gui_instance_data,
          "Warning: unprintable characters replaced with '.', only CLI mode supports binary files.",
          "", 1);

    /* load file in text box */
    GtkTextBuffer *buffer;
    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_text));
    gtk_text_buffer_set_text(buffer, gui_instance_data.session_data,
        gui_instance_data.session_data_len);

    /* display filename in path box */
    char lab_text[128];
    sprintf(lab_text, "Session setup - %s", file_name);
    gtk_label_set_text(GTK_LABEL(gui_data.ses_set_lab), lab_text);

    g_free(file_name);
    free(gui_data.cur_folder);
    gui_data.cur_folder = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(dialog));
    }
  gtk_widget_destroy(dialog);

  }

void gui_file_save()
  {
  int i;
  FILE *save_file;
  GtkWidget *dialog;
  dialog = gtk_file_chooser_dialog_new("Save File", GTK_WINDOW(gui_data.window),
  GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
      GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
  gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(dialog), gui_data.cur_folder);
  gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(dialog), TRUE);

  if (gtk_dialog_run(GTK_DIALOG (dialog)) == GTK_RESPONSE_ACCEPT)
    {
    char *file_name;
    file_name = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(dialog));
    GtkTextIter start, end;
    GtkTextBuffer *buffer;
    buffer = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_text));
    gtk_text_buffer_get_bounds(buffer, &start, &end);
    char *gui_session_text = gtk_text_buffer_get_text(buffer, &start, &end,
        FALSE);

    if ((save_file = fopen(file_name, "wb")) == NULL)
    tj_error(&gui_instance_data, "Can not save session file ", file_name, 1);
    else
      {
      for (i = 0; i <= strlen(gui_session_text) - 1; i++)
      putc(gui_session_text[i], save_file);
      fclose(save_file);
      tj_error(&gui_instance_data, "File saved", "", 0);
      }

    /* display filename in path box */
    char lab_text[128];
    sprintf(lab_text, "Session setup - %s", file_name);
    gtk_label_set_text(GTK_LABEL(gui_data.ses_set_lab), lab_text);

    g_free(file_name);
    free(gui_data.cur_folder);
    gui_data.cur_folder = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(dialog));

    }
  gtk_widget_destroy(dialog);

  }

static void ssl_check_toggled(GtkToggleButton *ssl_checkbox)
  {
  if (gtk_toggle_button_get_active(ssl_checkbox))
    gui_instance_data.use_ssl = 1;
  else
    gui_instance_data.use_ssl = 0;
  }

static void wait_data_check_toggled(GtkToggleButton *wait_data_checkbox)
  {
  if (gtk_toggle_button_get_active(wait_data_checkbox))
    gui_instance_data.wait_conn = 1;
  else
    gui_instance_data.wait_conn = 0;
  }

static void close_last_check_toggled(GtkToggleButton *close_last_checkbox)
  {
  if (gtk_toggle_button_get_active(close_last_checkbox))
    gui_instance_data.no_wait_last = 1;
  else
    gui_instance_data.no_wait_last = 0;
  }

static void auto_scroll_toggled(GtkToggleButton *auto_scr_checkbox)
  {
  if (gtk_toggle_button_get_active(auto_scr_checkbox))
    gui_data.auto_scroll = 1;
  else
    gui_data.auto_scroll = 0;
  }

static void auto_clearstat_toggled(GtkToggleButton *auto_clearstat_checkbox)
  {
  if (gtk_toggle_button_get_active(auto_clearstat_checkbox))
    gui_data.auto_clearstat = 1;
  else
    gui_data.auto_clearstat = 0;
  }

void gui_init(int argc, char *argv[], struct tj_data instance_data)
  {
  /* INIT */
  srand((unsigned int) (time(NULL)));
  gtk_init(&argc, &argv);

  gui_instance_data = instance_data;
  gui_instance_data.gui = 1;
  gui_instance_data.verbose = 1;
  gui_instance_data.debug = 1;
  gui_data.auto_scroll = 0;
  gui_data.listening = 0;
  gui_data.auto_clearstat = 1;
  gui_data.cur_folder = getcwd(NULL, 0);

  /* window */
  gui_data.window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW (gui_data.window), "Tcpjunk");
  gtk_container_set_border_width(GTK_CONTAINER (gui_data.window), 1);
  g_signal_connect(G_OBJECT(gui_data.window), "delete_event", G_CALLBACK(
          delete_event), NULL);
  gtk_window_set_icon_from_file(GTK_WINDOW(gui_data.window), "/usr/share/pixmaps/tcpjunk.png", NULL);
  gtk_window_resize(GTK_WINDOW (gui_data.window), 640, 480);

  /* boxes */
  GtkWidget *main_box, *right_box, *conf_box, *session_set_box,
      *session_log_box, *status_box;
  main_box = gtk_hbox_new(FALSE, 0);
  right_box = gtk_vbox_new(FALSE, 0);
  conf_box = gtk_vbox_new(FALSE, 0);
  session_set_box = gtk_vbox_new(FALSE, 0);
  session_log_box = gtk_vbox_new(FALSE, 0);
  status_box = gtk_vbox_new(FALSE, 0);

  /* check boxes */
  GtkWidget *ssl_checkbox, *wait_data_checkbox, *close_last_checkbox,
      *auto_scr_checkbox, *auto_clearstat_checkbox;
  ssl_checkbox = gtk_check_button_new_with_label("Use SSL");
  wait_data_checkbox
      = gtk_check_button_new_with_label("Wait for data at first");
  close_last_checkbox = gtk_check_button_new_with_label(
      "Don't wait for data at last");
  auto_scr_checkbox
      = gtk_check_button_new_with_label("Auto scroll session log");
  auto_clearstat_checkbox
      = gtk_check_button_new_with_label("Auto clear status");
  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(auto_clearstat_checkbox), TRUE);
  g_signal_connect(G_OBJECT(ssl_checkbox), "toggled", G_CALLBACK(
          ssl_check_toggled), NULL);
  g_signal_connect(G_OBJECT(wait_data_checkbox), "toggled", G_CALLBACK(
          wait_data_check_toggled), NULL);
  g_signal_connect(G_OBJECT(close_last_checkbox), "toggled", G_CALLBACK(
          close_last_check_toggled), NULL);
  g_signal_connect(G_OBJECT(auto_scr_checkbox), "toggled", G_CALLBACK(
          auto_scroll_toggled), NULL);
  g_signal_connect(G_OBJECT(auto_clearstat_checkbox), "toggled", G_CALLBACK(
          auto_clearstat_toggled), NULL);

  /* textviews */
  gui_data.session_text = gtk_text_view_new();
  gui_data.session_log_text = gtk_text_view_new();
  gui_data.status_text = gtk_text_view_new();
  gtk_text_view_set_editable(GTK_TEXT_VIEW(gui_data.session_log_text), FALSE);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(gui_data.status_text), FALSE);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui_data.session_text),
      GTK_WRAP_CHAR);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui_data.session_log_text),
      GTK_WRAP_CHAR);
  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(gui_data.status_text),
      GTK_WRAP_WORD);
  g_signal_connect(GTK_TEXT_VIEW(gui_data.session_text), "copy-clipboard",
      G_CALLBACK(gui_show_selection), NULL);
  g_signal_connect(GTK_TEXT_VIEW(gui_data.session_log_text), "copy-clipboard",
      G_CALLBACK(gui_show_selection), NULL);

  /* tags */
  GtkTextBuffer *ses_txt_buf, *ses_log_buf, *ses_stat_buf;
  ses_txt_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_text));
  ses_log_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.session_log_text));
  ses_stat_buf = gtk_text_view_get_buffer(GTK_TEXT_VIEW(gui_data.status_text));
  gtk_text_buffer_create_tag(ses_txt_buf, "yellow_hl", "background", "#f1f3cd", NULL);
  gtk_text_buffer_create_tag(ses_log_buf, "red_fg", "foreground", "#991111", "background", "#fff8f8",NULL);
  gtk_text_buffer_create_tag(ses_log_buf, "blue_fg", "foreground", "#00007f", "background", "#f6f6ff", NULL);
  gtk_text_buffer_create_tag(ses_log_buf, "yellow_hl", "background", "#f1f3cd", NULL);
  gtk_text_buffer_create_tag(ses_stat_buf, "red_fg", "foreground", "#ff0000", NULL);

  /* textviews fonts */
  PangoFontDescription *font_desc;
  font_desc = pango_font_description_from_string("Fixed 13");
  gtk_widget_modify_font(gui_data.session_text, font_desc);
  gtk_widget_modify_font(gui_data.session_log_text, font_desc);
  pango_font_description_free(font_desc);

  /* text entries */
  gui_data.host_entry = gtk_entry_new();
  gui_data.port_entry = gtk_entry_new();
  gui_data.init_to_entry = gtk_entry_new();
  gui_data.sock_to_entry = gtk_entry_new();
  gui_data.repeat_entry = gtk_entry_new();
  gui_data.syntax_entry = gtk_entry_new();
  gtk_entry_set_text(GTK_ENTRY(gui_data.init_to_entry), GUI_RDELAY);
  gtk_entry_set_text(GTK_ENTRY(gui_data.sock_to_entry), GUI_SDELAY);
  gtk_entry_set_text(GTK_ENTRY(gui_data.repeat_entry), "1");
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.init_to_entry), 2);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.sock_to_entry), 5);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.port_entry), 5);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.host_entry), 16);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.repeat_entry), 4);
  gtk_entry_set_width_chars(GTK_ENTRY(gui_data.syntax_entry), 14);
  g_signal_connect(G_OBJECT(gui_data.syntax_entry), "key-press-event", G_CALLBACK(
          syntax_entry_key_press), NULL);

  /* labels */
  GtkWidget *host_lab, *port_lab, *sect_beg_lab, *sect_end_lab, *sock_to_lab,
  *init_to_lab, *ses_log_lab, *status_lab, *repeat_lab, *syntax_lab;
  host_lab = gtk_label_new("Hostname/IP");
  port_lab = gtk_label_new("Port");
  sect_beg_lab = gtk_label_new("Section begins token ");
  sect_end_lab = gtk_label_new("Section ends token ");
  init_to_lab = gtk_label_new("Initial timeout (sec) ");
  sock_to_lab = gtk_label_new("Reply timeout (ms) ");
  gui_data.ses_set_lab = gtk_label_new("Session setup");
  ses_log_lab = gtk_label_new("Session log");
  status_lab = gtk_label_new("Status");
  repeat_lab = gtk_label_new("Repeat session");
  syntax_lab = gtk_label_new("Highlight");
  gtk_misc_set_alignment( GTK_MISC(host_lab), 0.0, 1.0 );
  gtk_misc_set_alignment( GTK_MISC(port_lab), 0.0, 1.0 );
  gtk_misc_set_alignment( GTK_MISC(sect_beg_lab), 1.0, 0.0 );
  gtk_misc_set_alignment( GTK_MISC(sect_end_lab), 1.0, 0.0 );
  gtk_misc_set_alignment( GTK_MISC(init_to_lab), 0.0, 1.0 );
  gtk_misc_set_alignment( GTK_MISC(sock_to_lab), 0.0, 1.0 );
  gtk_misc_set_alignment( GTK_MISC(repeat_lab), 0.0, 1.0 );
  gtk_misc_set_alignment( GTK_MISC(gui_data.ses_set_lab), 0.0, 1.0 );
  gtk_misc_set_alignment( GTK_MISC(ses_log_lab), 0.0, 1.0 );
  gtk_misc_set_alignment( GTK_MISC(status_lab), 0.0, 1.0 );
  gtk_misc_set_alignment( GTK_MISC(syntax_lab), 0.0, 1.0 );

  /* tables */
  GtkWidget *conf_table, *file_table, *other_table, *actions_table;
  conf_table = gtk_table_new (12, 2, FALSE);
  file_table = gtk_table_new (2, 2, FALSE);
  other_table = gtk_table_new (3, 2, FALSE);
  actions_table = gtk_table_new (2, 1, FALSE);

  /* frames */
  GtkWidget *session_set_frame, *session_log_frame, *conf_frame,
  *file_frame, *status_frame, *other_frame, *actions_frame;
  session_set_frame = gtk_frame_new(NULL);
  session_log_frame = gtk_frame_new(NULL);
  conf_frame = gtk_frame_new(" Connection setup ");
  file_frame = gtk_frame_new(" Session file ");
  status_frame = gtk_frame_new(NULL);
  other_frame = gtk_frame_new(" GUI options ");
  actions_frame = gtk_frame_new(" Actions ");

  /* scroll windows */
  GtkWidget *ses_set_scr_win, *session_log_scr_win, *status_scr_win;
  ses_set_scr_win = gtk_scrolled_window_new (NULL, NULL);
  session_log_scr_win = gtk_scrolled_window_new (NULL, NULL);
  status_scr_win = gtk_scrolled_window_new (NULL, NULL);
  gtk_container_add (GTK_CONTAINER (ses_set_scr_win), gui_data.session_text);
  gtk_container_add (GTK_CONTAINER (session_log_scr_win), gui_data.session_log_text);
  gtk_container_add (GTK_CONTAINER (status_scr_win), gui_data.status_text);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (ses_set_scr_win),
      GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (session_log_scr_win),
      GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (status_scr_win),
      GTK_POLICY_AUTOMATIC, GTK_POLICY_ALWAYS);

  /* buttons */
  GtkWidget *save_butt;
  gui_data.connect_butt = gtk_button_new_with_label ("Connect [Ctrl+S]");
  gui_data.listen_butt = gtk_button_new_with_label ("Listen");
  save_butt = gtk_button_new_with_label ("Save as");
  gui_data.open_butt = gtk_button_new_with_label ("Open");
  g_signal_connect_swapped (G_OBJECT (gui_data.connect_butt), "clicked",
      G_CALLBACK (gui_connect), (gpointer) gui_data.connect_butt);
  g_signal_connect_swapped (G_OBJECT (gui_data.listen_butt), "clicked",
      G_CALLBACK (gui_listen), (gpointer) gui_data.listen_butt);
  g_signal_connect_swapped (G_OBJECT (save_butt), "clicked",
      G_CALLBACK (gui_file_save), NULL);
  g_signal_connect_swapped (G_OBJECT (gui_data.open_butt), "clicked",
      G_CALLBACK (gui_file_open), NULL);

  /* alignments */
  GtkWidget *port_alig, *init_to_alig, *sock_to_alig, *conf_table_alig,
  *file_table_alig, *other_table_alig, *host_alig, *repeat_alig, *actions_table_alig,
      *syntax_alig;
      port_alig = gtk_alignment_new (0, 0, 0, 0);
      host_alig = gtk_alignment_new (0, 0, 0, 0);
      init_to_alig = gtk_alignment_new (0, 0, 0, 0);
      sock_to_alig = gtk_alignment_new (0, 0, 0, 0);
      repeat_alig = gtk_alignment_new (0, 0, 0, 0);
      conf_table_alig = gtk_alignment_new (0, 0, 0, 0);
      file_table_alig = gtk_alignment_new (1, 1, 1, 0);
      other_table_alig = gtk_alignment_new (1, 1, 1, 0);
      actions_table_alig = gtk_alignment_new (0.5, 0, 1, 0);
      syntax_alig = gtk_alignment_new (0.5, 0, 1, 0);
      gtk_container_add (GTK_CONTAINER (port_alig), gui_data.port_entry);
      gtk_container_add (GTK_CONTAINER (host_alig), gui_data.host_entry);
      gtk_container_add (GTK_CONTAINER (init_to_alig), gui_data.init_to_entry);
      gtk_container_add (GTK_CONTAINER (sock_to_alig), gui_data.sock_to_entry);
      gtk_container_add (GTK_CONTAINER (conf_table_alig), conf_table);
      gtk_container_add (GTK_CONTAINER (file_table_alig), file_table);
      gtk_container_add (GTK_CONTAINER (actions_table_alig), actions_table);
      gtk_container_add (GTK_CONTAINER (other_table_alig), other_table);
      gtk_container_add (GTK_CONTAINER (repeat_alig), gui_data.repeat_entry);
      gtk_container_add (GTK_CONTAINER (syntax_alig), gui_data.syntax_entry);
      gtk_alignment_set_padding(GTK_ALIGNMENT(conf_table_alig), 3, 3, 5, 5);
      gtk_alignment_set_padding(GTK_ALIGNMENT(file_table_alig), 3, 3, 5, 5);
      gtk_alignment_set_padding(GTK_ALIGNMENT(other_table_alig), 3, 3, 5, 5);
      gtk_alignment_set_padding(GTK_ALIGNMENT(actions_table_alig), 3, 3, 5, 5);
      gtk_alignment_set_padding(GTK_ALIGNMENT(syntax_alig), 3, 3, 5, 5);

      /* vpaneds */
      GtkWidget *sess_pan, *stat_pan;
      sess_pan = gtk_vpaned_new();
      stat_pan = gtk_vpaned_new();

      gtk_paned_set_position(GTK_PANED(stat_pan), 400);
      gtk_paned_set_position(GTK_PANED(sess_pan), 200);

      /* pack sessions texts */
      gtk_container_add (GTK_CONTAINER (session_set_frame), ses_set_scr_win);
      gtk_container_add (GTK_CONTAINER (session_log_frame), session_log_scr_win);
      gtk_container_add (GTK_CONTAINER (conf_frame), conf_table_alig);
      gtk_box_pack_start(GTK_BOX(session_set_box), GTK_WIDGET(gui_data.ses_set_lab),
          FALSE, FALSE, 1);
      gtk_box_pack_start(GTK_BOX(session_set_box), GTK_WIDGET(session_set_frame),
          TRUE, TRUE, 1);
      gtk_box_pack_start(GTK_BOX(session_log_box), GTK_WIDGET(ses_log_lab),
          FALSE, FALSE, 1);
      gtk_box_pack_start(GTK_BOX(session_log_box), GTK_WIDGET(session_log_frame),
          TRUE, TRUE, 1);
      gtk_container_add (GTK_CONTAINER (status_frame), status_scr_win);
      gtk_box_pack_start(GTK_BOX(status_box), GTK_WIDGET(status_lab),
                FALSE, FALSE, 1);
      gtk_box_pack_start(GTK_BOX(status_box), GTK_WIDGET(status_frame),
                TRUE, TRUE, 1);

      /* fill conf table */
      gtk_table_attach( GTK_TABLE(conf_table), host_lab,
          0, 1, 0, 1, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2 );
      gtk_table_attach( GTK_TABLE(conf_table), port_lab,
          1, 2, 0, 1, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 0 );
      gtk_table_attach (GTK_TABLE (conf_table), host_alig,
          0, 1, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), port_alig,
          1, 2, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), ssl_checkbox,
          0, 2, 3, 4, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), wait_data_checkbox,
          0, 2, 4, 5, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), close_last_checkbox,
          0, 2, 5, 6, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), init_to_lab,
          0, 1, 6, 7, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), init_to_alig,
          1, 2, 6, 7, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), sock_to_lab,
          0, 1, 7, 8, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), sock_to_alig,
          1, 2, 7, 8, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), repeat_lab,
          0, 1, 8, 9, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (conf_table), repeat_alig,
          1, 2, 8, 9, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);

      /* pack file table */
      gtk_table_attach( GTK_TABLE(file_table), gui_data.open_butt,
          0, 1, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2 );
      gtk_table_attach( GTK_TABLE(file_table), save_butt,
          1, 2, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2 );
      gtk_container_add (GTK_CONTAINER (file_frame), file_table_alig);

      /* pack other table */
      gtk_table_attach (GTK_TABLE (other_table), auto_scr_checkbox,
          0, 2, 0, 1, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (other_table), auto_clearstat_checkbox,
          0, 2, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (other_table), syntax_lab,
          0, 1, 2, 3, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (other_table), syntax_alig,
          1, 2, 2, 3, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_container_add (GTK_CONTAINER (other_frame), other_table_alig);

      /* pack actions table */
      gtk_table_attach (GTK_TABLE (actions_table), gui_data.connect_butt,
          0, 1, 0, 1, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_table_attach (GTK_TABLE (actions_table), gui_data.listen_butt,
          0, 1, 1, 2, GTK_EXPAND|GTK_FILL, GTK_SHRINK, 0, 2);
      gtk_container_add (GTK_CONTAINER (actions_frame), actions_table_alig);

      /* accelerators */
      GtkAccelGroup *ac_grp;
      ac_grp = gtk_accel_group_new();
      gtk_window_add_accel_group(GTK_WINDOW(gui_data.window), ac_grp);
      gtk_widget_add_accelerator (gui_data.connect_butt, "activate", ac_grp, GDK_S,
          GDK_CONTROL_MASK, GTK_ACCEL_VISIBLE);

      /* pack conf box */
      gtk_box_pack_start(GTK_BOX(conf_box), GTK_WIDGET(file_frame), FALSE, FALSE, 1);
      gtk_box_pack_start(GTK_BOX(conf_box), GTK_WIDGET(conf_frame), FALSE, FALSE, 1);
      gtk_box_pack_start(GTK_BOX(conf_box), GTK_WIDGET(other_frame), FALSE, FALSE, 1);
      gtk_box_pack_start(GTK_BOX(conf_box), GTK_WIDGET(actions_frame), FALSE, FALSE, 1);

      /* pack right box */
      gtk_paned_add1(GTK_PANED(sess_pan), session_set_box);
      gtk_paned_add2(GTK_PANED(sess_pan), session_log_box);
      gtk_paned_add1(GTK_PANED(stat_pan), sess_pan);
      gtk_paned_add2(GTK_PANED(stat_pan), status_box);
      gtk_box_pack_start(GTK_BOX(right_box), GTK_WIDGET(stat_pan), TRUE, TRUE, 1);

      /* pack window */
      gtk_box_pack_start(GTK_BOX(main_box), GTK_WIDGET(conf_box), FALSE, FALSE, 3);
      gtk_box_pack_start(GTK_BOX(main_box), GTK_WIDGET(right_box), TRUE, TRUE, 1);
      gtk_container_add (GTK_CONTAINER (gui_data.window), main_box);
      gtk_widget_show_all (gui_data.window);

      /* show version */
      tj_error(&gui_instance_data, "Tcpjunk ver ", VER, 0);

      gtk_main ();

      }

