/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <fcntl.h>
#include "tcpjunk.h"

void junk(struct tj_data instance_data)
  {

  /* setup session send buffer */
  int send_buf_len = instance_data.session_data_len;
  char *send_buf;
  send_buf = tj_malloc(send_buf, send_buf_len);

  /* other stuff */
  int buf_id;
  int loop_counter = instance_data.loop;
  int keepalive_counter = instance_data.keepalive;
  int rc;
  int cl_socket;
  int r_timeout;

  /* socket stuff */
  struct sockaddr_in sa;
  memset(&sa, 0, sizeof(sa));
  sa.sin_family = AF_INET;
  sa.sin_port = htons(instance_data.dest_port);
  memcpy(&sa.sin_addr.s_addr, instance_data.dest_addr,
      instance_data.dest_addr_len);

  /* SSL Stuff */
  int err;
  SSL_CTX *ctx = NULL;
  SSL *ssl = NULL;
  SSL_METHOD *meth;

  /* Prepare SSL Stuff if needed */
  if (instance_data.use_ssl == 1)
    {
    SSL_library_init();
    SSL_load_error_strings();

    /* Create SSL_CTX structure */
    meth = SSLv3_method();
    ctx = SSL_CTX_new(meth);
    if ((ctx) == NULL)
      {
      tj_error(&instance_data, "Cannot create an SSL_CTX structure", "", 1);
      exit(EXIT_FAILURE);
      }
    /* SSL structure */
    ssl = SSL_new(ctx);
    if ((ssl) == NULL)
      {
      tj_error(&instance_data, "Cannot create an SSL structure", "", 1);
      exit(EXIT_FAILURE);
      }
    }

  /**
   ***  main loop
   **/
  while (instance_data.loop == 0 ? 1 : loop_counter--)
    {

    /* load random session data file */
    if (instance_data.rand_file == 1)
      {
      /* choose random file */
      set_random_session_file(&instance_data);
      /* load session file */
      load_session_data(&instance_data);
      /* check config */
      find_sections(&instance_data);
      }

    /* allocate a socket */
    cl_socket = socket(AF_INET,SOCK_STREAM, 0);
    if (cl_socket < 0)
      tj_error(&instance_data, "Socket allocation failed: ", strerror(errno), 1);

    /* bind to different source IP address */
    if (strcmp(instance_data.interface, "") != 0)
      {
      struct sockaddr_in ba;
      memset(&ba, 0, sizeof(ba));
      ba.sin_family = AF_INET;
      char bind_if_name[32];
      char new_src_ip[32];

      /* setup alias */
      sprintf(bind_if_name, "%s:%d", instance_data.interface,
          instance_data.instance_num + instance_data.offset - 1);

      if (instance_data.rand_ip == 1)
        /* set random IP */
        sprintf(new_src_ip, "%d.%d.%d.%d", rand() % 224, rand() % 224, rand()
            % 224, rand() % 224);
      else
        /* setup specific IP */
        sprintf(new_src_ip, instance_data.src_ip);

      /* make sure address is cool */
      check_address(&instance_data, new_src_ip);

      /* setup IP on alias */
      set_src_ip(&instance_data, bind_if_name, new_src_ip);

      /* setup source port if needed */
      if (instance_data.src_port > -1)
        ba.sin_port = ntohs(instance_data.src_port);

      /* bind socket to IP address */
      memcpy(&ba.sin_addr.s_addr, instance_data.src_addr,
          instance_data.src_addr_len);
      if (bind(cl_socket, (struct sockaddr *) &ba, sizeof(struct sockaddr_in))
          < 0)
        tj_error(&instance_data, "Socket binding failed: ", strerror(errno), 1);
      }
    else if (instance_data.src_port > -1)
      {
      /* only bind to different source port */
      struct sockaddr_in ba;
      memset(&ba, 0, sizeof(ba));
      ba.sin_family = AF_INET;
      sa.sin_addr.s_addr = INADDR_ANY;
      ba.sin_port = ntohs(instance_data.src_port);

      /* bind socket to IP address */
      memcpy(&ba.sin_addr.s_addr, instance_data.src_addr,
          instance_data.src_addr_len);
      if (bind(cl_socket, (struct sockaddr *) &ba, sizeof(struct sockaddr_in))
          < 0)
        tj_error(&instance_data, "Socket binding failed: ", strerror(errno), 1);
      }

    /* connect */
    rc = tj_connect(cl_socket, sa, &instance_data);

    /* if no error, start transfers */
    if (rc == 0)
      {
      if (instance_data.use_ssl == 1)
        {
        /* Assign socket to SSL structure */
        SSL_set_fd(ssl, cl_socket);

        /* SSL Handshake */
        err = SSL_connect(ssl);
        if ((err) == -1)
          tj_error(&instance_data, "SSL connect ", ERR_error_string(
              ERR_get_error(), NULL), 1);
        }
      send_data:

      buf_id = 0;
      r_timeout = 0;
      /* get next send buffer from config */
      while (buf_id < instance_data.sect_idx)
        {
        /* see what client has to say on start */
        if (buf_id == 0 && instance_data.wait_conn == 1)
          r_timeout = read_reply(cl_socket, ssl, &instance_data);
        if (r_timeout > 0)
          {
          tj_error(&instance_data, "Reply timeout", "", 1);
          goto close_con;
          }

        /* prepare a temp send buffer */
        if (send_buf_len != instance_data.sections_len[buf_id])
          {
          send_buf_len = instance_data.sections_len[buf_id];
          send_buf = tj_realloc(send_buf, send_buf_len);
          }

        memset(send_buf, 0, send_buf_len);
        memcpy(send_buf, instance_data.sections[buf_id], send_buf_len);

        /* process instance tags */
        send_buf_len = instance_num_data(&send_buf, &instance_data,
            send_buf_len);

        /* process system tags */
        send_buf_len = system_data(&send_buf, &instance_data, send_buf_len);

        /* process char tags */
        send_buf_len = char_data(&send_buf, &instance_data, send_buf_len);

        /* process repeat tags */
        send_buf_len = repeat_data(&send_buf, &instance_data, send_buf_len);

        /* process fuzz tags */
        send_buf_len = fuzz_data(&send_buf, &instance_data, send_buf_len);

        /* send data*/
        send_data(cl_socket, ssl, &instance_data, send_buf, send_buf_len);

        /* see what server has to say, if we need to */
        if (!(instance_data.no_wait_last == 1 && buf_id
            == instance_data.sect_idx - 1))
          r_timeout = read_reply(cl_socket, ssl, &instance_data);
        if (r_timeout > 0)
          {
          tj_error(&instance_data, "Reply timeout", "", 1);
          goto close_con;
          }

        buf_id++;
        }

      /* if keep alive send again */
      if (keepalive_counter > 1 && r_timeout == 0)
        {
        keepalive_counter--;
        goto send_data;
        }

      /* close socket */
      close_con:

      tj_disconnect(cl_socket, &instance_data, ssl, ctx);

      if (instance_data.debug == 1)
        goto exit_child;
      }
    else
      {
      /* connection failed */
      tj_error(&instance_data, "Connection failed: ", strerror(errno), 1);
      tj_disconnect(cl_socket, &instance_data, ssl, ctx);
      }

    /* wait if need to */
    if (instance_data.wait > 0)
      usleep(instance_data.wait);

    /* stop if only one instance debug mode */
    if (instance_data.debug == 1)
      break;

    } /* end of main loop */
  exit_child:

  if (instance_data.use_ssl == 1)
    {
    err = SSL_shutdown(ssl);
    if ((err) == -1)
      tj_error(&instance_data, "SSL Disconnect ", ERR_error_string(
          ERR_get_error(), NULL), 1);
    SSL_free(ssl);
    SSL_CTX_free(ctx);
    }

  free(send_buf);
  if (instance_data.gui == 0)
    exit(EXIT_SUCCESS);
  else
    return;

  }

