/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <limits.h>
#include "tcpjunk.h"

int main(int argc, char *argv[])
  {

  /* we shall ignore sigpipes */
  signal(SIGPIPE, SIG_IGN);

  /* clean dead processes */
  struct sigaction sig_a;
  sig_a.sa_handler = sigchld_handler;
  sigemptyset(&sig_a.sa_mask);
  sig_a.sa_flags = SA_RESTART;
  sigaction(SIGCHLD, &sig_a, NULL);

  /* some vars */
  struct tj_data instance_data;
  int arg_num = 0;
  int children_num = 0;
  char srv_addr [FILENAME_MAX] = "";
  int server_mode = 0;
  instance_data.wait = 0;
  instance_data.itimeout = 2;
  instance_data.stimeout = 100;
  instance_data.stimeout *= 1000;
  instance_data.keepalive = 1;
  instance_data.loop = 0;
  instance_data.wait_conn = 0;
  instance_data.no_wait_last = 0;
  instance_data.tcp_opt = 0;
  instance_data.use_ssl = 0;
  strcpy(instance_data.server_cert , SERVER_CERT);
  strcpy(instance_data.server_key , SERVER_KEY);
  instance_data.gui = 0;
  strcpy(instance_data.interface, "");
  strcpy(instance_data.src_ip, "");
  instance_data.offset = 0;
  instance_data.segment_delay = 0;
  instance_data.rand_ip = 0;
  instance_data.segment_size = 0;
  instance_data.verbose = 0;
  instance_data.rand_file = 0;
  instance_data.src_port = -1;
  instance_data.instance_num = 1;

  /**
   *** parse params
   **/
  if (argc < 2)
    {
    show_help();
    exit(EXIT_SUCCESS);
    }

  for (arg_num = 1; arg_num < argc; arg_num++)
    {

    /* GUI mode */
    if (strcmp(argv[arg_num], "-x") == 0)
      {
      gui_init(argc, argv, instance_data);
      exit(EXIT_SUCCESS);
      }

    /* help */
    if (strcmp(argv[arg_num], "-h") == 0)
      {
      show_help();
      exit(EXIT_SUCCESS);
      }

    /* server mode */
    if (strcmp(argv[arg_num], "-r") == 0)
      {
      server_mode = 1;
      continue;
      }

    /* destination address */
    if (strcmp(argv[arg_num], "-s") == 0 && arg_num < argc - 1)
      {
      strcpy(srv_addr, argv[arg_num + 1]);
      continue;
      }

    /* destination port */
    if (strcmp(argv[arg_num], "-p") == 0 && arg_num < argc - 1)
      {
      instance_data.dest_port = atoi(argv[arg_num + 1]);
      continue;
      }

    /* source port */
    if (strcmp(argv[arg_num], "-O") == 0 && arg_num < argc - 1)
      {
      instance_data.src_port = atoi(argv[arg_num + 1]);
      continue;
      }

    /* wait between rounds */
    if (strcmp(argv[arg_num], "-w") == 0 && arg_num < argc - 1)
      {
      instance_data.wait = atoi(argv[arg_num + 1]);
      instance_data.wait *= 1000;
      continue;
      }

    /* force debug */
    if (strcmp(argv[arg_num], "-d") == 0)
      {
      instance_data.debug = 2;
      continue;
      }

    /* config file */
    if (strcmp(argv[arg_num], "-c") == 0 && arg_num < argc - 1)
      {
      strcpy(instance_data.session_file_name, argv[arg_num + 1]);
      continue;
      }

    /* number of children */
    if (strcmp(argv[arg_num], "-i") == 0 && arg_num < argc - 1)
      {
      children_num = atoi(argv[arg_num + 1]);
      continue;
      }

    /* initial read timeout */
    if (strcmp(argv[arg_num], "-t") == 0 && arg_num < argc - 1)
      {
      instance_data.itimeout = atoi(argv[arg_num + 1]);
      continue;
      }

    /* socket read timeout */
    if (strcmp(argv[arg_num], "-T") == 0 && arg_num < argc - 1)
      {
      instance_data.stimeout = atoi(argv[arg_num + 1]);
      instance_data.stimeout *= 1000;
      continue;
      }

    /* looping */
    if (strcmp(argv[arg_num], "-l") == 0 && arg_num < argc - 1)
      {
      instance_data.loop = atoi(argv[arg_num + 1]);
      continue;
      }

    /* keepalive */
    if (strcmp(argv[arg_num], "-k") == 0 && arg_num < argc - 1)
      {
      instance_data.keepalive = atoi(argv[arg_num + 1]);
      continue;
      }

    /* wait on connection */
    if (strcmp(argv[arg_num], "-W") == 0)
      {
      instance_data.wait_conn = 1;
      continue;
      }

    /* dont wait after last section */
    if (strcmp(argv[arg_num], "-C") == 0)
      {
      instance_data.no_wait_last = 1;
      continue;
      }

    /* wait on connection */
    if (strcmp(argv[arg_num], "-v") == 0)
      {
      instance_data.verbose = 1;
      continue;
      }

    /* tcp optimizations */
    if (strcmp(argv[arg_num], "-u") == 0)
      {
      instance_data.tcp_opt = 1;
      continue;
      }

    /* use SSL */
    if (strcmp(argv[arg_num], "-S") == 0)
      {
      instance_data.use_ssl = 1;
      continue;
      }

    /* server cert */
    if (strcmp(argv[arg_num], "-E") == 0 && arg_num < argc - 1)
      {
      strcpy(instance_data.server_cert, argv[arg_num + 1]);
      continue;
      }

    /* server key */
    if (strcmp(argv[arg_num], "-K") == 0 && arg_num < argc - 1)
      {
      strcpy(instance_data.server_key, argv[arg_num + 1]);
      continue;
      }

    /* random  IP */
    if (strcmp(argv[arg_num], "-R") == 0)
      {
      instance_data.rand_ip = 1;
      continue;
      }

    /* source IP */
    if (strcmp(argv[arg_num], "-P") == 0 && arg_num < argc - 1)
      {
      strcpy(instance_data.src_ip, argv[arg_num + 1]);
      continue;
      }

    /* interface */
    if (strcmp(argv[arg_num], "-I") == 0 && arg_num < argc - 1)
      {
      strcpy(instance_data.interface, argv[arg_num + 1]);
      continue;
      }

    /* alias offset */
    if (strcmp(argv[arg_num], "-o") == 0 && arg_num < argc - 1)
      {
      instance_data.offset = atoi(argv[arg_num + 1]);
      continue;
      }

    /* segment size */
    if (strcmp(argv[arg_num], "-m") == 0 && arg_num < argc - 1)
      {
      instance_data.segment_size = atoi(argv[arg_num + 1]);
      if (instance_data.segment_size> SOCK_BUF_LEN)
      instance_data.segment_size = SOCK_BUF_LEN;
      continue;
      }

    /* wait before segment */
    if (strcmp(argv[arg_num], "-D") == 0 && arg_num < argc - 1)
      {
      instance_data.segment_delay = atoi(argv[arg_num + 1]);
      instance_data.segment_delay *= 1000;
      continue;
      }

    /* random session file */
    if (strcmp(argv[arg_num], "-f") == 0)
      {
      instance_data.rand_file = 1;
      continue;
      }

    }
  /**
   *** end of parameters check
   **/

  /* client mode checks */
  if (server_mode == 0)
    {
    /* looping */
    if (instance_data.loop> 0 && children_num == 0)
      {
      printf("tcpjunk: looping only relevant when using multiple instances\r\n");
      exit(EXIT_FAILURE);
      }

    /* number of children and debug */
    if (children_num == 0)
      {
      children_num = 1;
      instance_data.debug = 1;
      }
    else
      {
      if (instance_data.debug != 2)
      instance_data.debug = 0;
      if (children_num < 0 || children_num> 1000)
        {
        printf("tcpjunk: invalid number of instances \'%d\'\r\n", children_num);
        exit(EXIT_FAILURE);
        }
      }
    /* source IP */
    if (strcmp(instance_data.interface, ""))
      {
      if (instance_data.rand_ip == 0 && strcmp(instance_data.src_ip, "") == 0)
        {
        printf("tcpjunk: source IP (or -R) must be specified for -I option\n");
        exit(EXIT_FAILURE);
        }
      }

    }
  /* server mode checks */
  if (server_mode == 1)
  printf("tcpjunk: starting server mode, use -v/-d to show errors/data\n");

  /* check destination address */
  if (server_mode == 0)
  check_address(&instance_data, srv_addr);

  /* check DEST port */
  check_port(instance_data.dest_port, &instance_data);

  /* check SRC port */
  if (instance_data.src_port> -1)
  check_port(instance_data.src_port, &instance_data);

  if (instance_data.rand_file == 1)
  load_rand_file_list(&instance_data);
  else
    {
    /* load session file */
    load_session_data(&instance_data);

    /* find sections */
    find_sections(&instance_data);
    }

  /* optimize TCP */
  if (instance_data.tcp_opt == 1)
    {
    system("ulimit -n 10240");
    system("echo 1 > /proc/sys/net/ipv4/tcp_tw_recycle");
    system("echo 1 > /proc/sys/net/ipv4/tcp_tw_reuse");
    }

  /* start client mode */
  if (server_mode == 0)
    {
    int ci;
    for (ci = 1; ci <= children_num; ci++)
      {
      instance_data.instance_num = ci;
      if(!fork())
        {
        srand((unsigned int)(time(NULL) ^ getpid()) + getppid());
        junk(instance_data);
        exit(EXIT_SUCCESS);
        }
      }
    signal(SIGINT, cleanup);
    wait(NULL);
    }
  else
    {
    /* start server mode */
    int yes = 1;
    int ret;
    int s;
    if (instance_data.debug != 2)
    instance_data.debug = 0;

    /* allocate listening socket */
    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0)
    tj_error(&instance_data, "Failed to allocate socket: ", strerror(
            errno), 1);

    struct sockaddr_in sa;
    memset(&sa, 0, sizeof(sa));
    struct sockaddr_in rsa;
    sa.sin_family = AF_INET;
    sa.sin_addr.s_addr = INADDR_ANY;
    sa.sin_port = htons(instance_data.dest_port);

    /* set socket to be reuseable */
    ret = setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int));
    if (ret < 0)
    tj_error(&instance_data, "Error setting socket options: ",
        strerror(errno), 1);

    /* bind socket */
    ret = bind(s, (struct sockaddr *)&sa, sizeof(struct sockaddr));
    if (ret < 0)
      {
      tj_error(&instance_data, "Failed to bind socket: ",
          strerror(errno), 1);
      exit(EXIT_FAILURE);
      }

    /* listen on socket */
    ret = listen(s, BACKLOG);
    if (ret < 0)
      {
      tj_error(&instance_data, "Failed to listen on socket: ", strerror(
              errno), 1);
      exit(EXIT_FAILURE);
      }

    /* fork loop */
    unsigned int sin_size = sizeof(struct sockaddr_in);

    while(1)
      {

      if((instance_data.new_sock = accept(s, (struct sockaddr *)&rsa, &sin_size)) == -1)
        {
        tj_error(&instance_data, "Failed accept on socket: ", strerror(
                errno), 1);
        continue;
        }
      else
      tj_error(&instance_data, "Connection accepted", "", 0);

      instance_data.instance_num++;
      if (instance_data.instance_num == INT_MAX)
      instance_data.instance_num = 1;
      /* create new server instance */
      if(!fork())
        {
        close(s);
        srand((unsigned int)(time(NULL) ^ getpid()) + getppid());
        sjunk(instance_data);
        exit(EXIT_SUCCESS);
        }

      /* parent code continues */
      close(instance_data.new_sock);

      }
    }

  return EXIT_SUCCESS;
  }

