/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netdb.h>
#include <unistd.h>
#include "tcpjunk.h"

void sjunk(struct tj_data instance_data)
  {
  /* some vars */
  int r_timeout;
  int buf_id;
  int send_buf_len = instance_data.session_data_len;
  char *send_buf;
  send_buf = malloc(send_buf_len);
  int keepalive_counter = instance_data.keepalive;

  /* SSL Stuff */
  int err;
  SSL_CTX *ctx = NULL;
  SSL *ssl = NULL;
  SSL_METHOD *meth;

  /* prepare SSL stuff if needed */
  if (instance_data.use_ssl == 1)
    {
    SSL_library_init();
    SSL_load_error_strings();

    /* SSL_METHOD structure */
    meth = SSLv3_method();
    ctx = SSL_CTX_new(meth);
    if ((ctx) == NULL)
      {
      printf("Cannot create an SSL_CTX structure\n");
      goto finish;
      }

    /* load server certificate */
    if (SSL_CTX_use_certificate_file(ctx, instance_data.server_cert,
        SSL_FILETYPE_PEM) <= 0)
      {
      tj_error(&instance_data, "SSL certificate ", ERR_error_string(
          ERR_get_error(), NULL), 1);
      goto finish;
      }

    /* load private-key */
    if (SSL_CTX_use_PrivateKey_file(ctx, instance_data.server_key,
        SSL_FILETYPE_PEM) <= 0)
      {
      tj_error(&instance_data, "SSL key ", ERR_error_string(ERR_get_error(),
          NULL), 1);
      goto finish;
      }

    /* check certificate and key match */
    if (!SSL_CTX_check_private_key(ctx))
      {
      tj_error(&instance_data, "SSL - ", "Key does not match certificate", 1);
      goto finish;
      }

    /* SSL structure */
    ssl = SSL_new(ctx);

    if ((ssl) == NULL)
      {
      tj_error(&instance_data, "Cannot create SSL structure", "", 1);
      goto finish;
      }

    /* Assign socket to SSL structure */
    SSL_set_fd(ssl, instance_data.new_sock);

    /* SSL Handshake */
    err = SSL_accept(ssl);
    if ((err) == -1)
      {
      tj_error(&instance_data, "SSL ", ERR_error_string(ERR_get_error(), NULL), 1);
      goto finish;
      }
    }

  /* load random session data file */
  if (instance_data.rand_file == 1)
    {
    /* choose random file */
    set_random_session_file(&instance_data);
    /* load session file */
    load_session_data(&instance_data);
    /* check config */
    find_sections(&instance_data);
    }

  /* start transfer */
  send_data:

  buf_id = 0;
  r_timeout = 0;

  /* get next send buffer from file */
  while (buf_id < instance_data.sect_idx)
    {
    /* see what client has to say on start */
    if (buf_id == 0 && instance_data.wait_conn == 1)
      r_timeout = read_reply(instance_data.new_sock, ssl, &instance_data);
    if (r_timeout > 0)
      {
      tj_error(&instance_data, "No data received, closing connection", "", 1);
      goto finish;
      }

    /* prepare a temp send buffer */
    if (send_buf_len != instance_data.sections_len[buf_id])
      {
      send_buf_len = instance_data.sections_len[buf_id];
      send_buf = tj_realloc(send_buf, send_buf_len);
      }

    memcpy(send_buf, instance_data.sections[buf_id], send_buf_len);

    /* process instance tags */
    send_buf_len = instance_num_data(&send_buf, &instance_data, send_buf_len);

    /* process system tags */
    send_buf_len = system_data(&send_buf, &instance_data, send_buf_len);

    /* process char tags */
    send_buf_len = char_data(&send_buf, &instance_data, send_buf_len);

    /* process repeat tags */
    send_buf_len = repeat_data(&send_buf, &instance_data, send_buf_len);

    /* process fuzz tags */
    send_buf_len = fuzz_data(&send_buf, &instance_data, send_buf_len);

    /* send data to client */
    send_data(instance_data.new_sock, ssl, &instance_data, send_buf,
        send_buf_len);

    /* see what client has to say, if we need to */
    if (!(instance_data.no_wait_last == 1 && buf_id == instance_data.sect_idx
        - 1))
      {
      r_timeout = read_reply(instance_data.new_sock, ssl, &instance_data);
      if (r_timeout > 0)
        {
        tj_error(&instance_data, "No data received, closing connection", "", 1);
        goto finish;
        }
      }

    buf_id++;
    }

  /* if keep alive send again */
  if (keepalive_counter > 1 && r_timeout == 0)
    {
    keepalive_counter--;
    goto send_data;
    }

  /* complete */
  finish:

  free(send_buf);
  tj_disconnect(instance_data.new_sock, &instance_data, ssl, ctx);
  if (instance_data.gui == 0)
    exit(EXIT_SUCCESS);
  else
    return;

  }

