/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/wait.h>
#include <netdb.h>
#include <sys/ioctl.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include "tcpjunk.h"

int system_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len)
  {
  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *system_tag = "<sys>";
  char *system_tag_end = "</sys>";
  char *system_tag_loc = NULL;

  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char system_cmd[FILENAME_MAX];
  char *system_cmd_loc;
  int tags_processed = 0;
  FILE *cmd_pipe;
  int pipe_char;
  int cmd_out_loc = 0, cmd_out_size = 1024;
  char *cmd_out = NULL;
  cmd_out = tj_malloc(cmd_out, cmd_out_size);

  while ((system_tag_loc = strstr(send_buf_loc, system_tag)))
    {
    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, system_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += system_tag_loc - send_buf_loc;

    /* advance send_buf_loc to end of tag */
    send_buf_loc = system_tag_loc + strlen(system_tag);

    /* point to start of tag data */
    system_cmd_loc = send_buf_loc;

    /* advance send_buf_loc to end of tag data */
    if (!(send_buf_loc = strstr(send_buf_loc, system_tag_end)))
      goto done;

    /* parse tag */
    strncpy(system_cmd, system_cmd_loc, send_buf_loc - system_cmd_loc);

    /* advance send_buf_loc to end of closing tag */
    send_buf_loc += strlen(system_tag_end);

    /* get stuff from the pipe */
    if ((cmd_pipe = popen(system_cmd, "r")) != NULL)
      {
      while ((pipe_char = getc(cmd_pipe)) != EOF)
        {
        cmd_out[cmd_out_loc++] = pipe_char;
        if (cmd_out_loc == cmd_out_size)
          {
          cmd_out_size *= 2;
          cmd_out = tj_realloc(cmd_out, cmd_out_size);
          }
        }
      /* resize prep_buf and add data */
      prep_buf_loc_offset = prep_buf_loc - prep_buf;
      prep_buf_len = prep_buf_len + cmd_out_loc;
      prep_buf = tj_realloc(prep_buf, prep_buf_len);
      prep_buf_loc = prep_buf + prep_buf_loc_offset;
      memcpy(prep_buf_loc, cmd_out, cmd_out_loc);
      prep_buf_loc += cmd_out_loc;
      }
    (void) pclose(cmd_pipe);

    tags_processed = 1;
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy prep buf to send buf */
    send_buf_len = prep_buf_len;
    send_buf = tj_realloc(send_buf, send_buf_len);
    memcpy(send_buf, prep_buf, send_buf_len);
    *send_buf_p = send_buf;
    }

  done:

  free(cmd_out);
  free(prep_buf);
  return send_buf_len;
  }

int repeat_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len)
  {
  char *send_buf = *send_buf_p;
  char *send_buf_loc;
  char *repeat_tag = "<rep ";
  char *repeat_tag_end = "</rep>";
  char *repeat_tag_loc = NULL;
  char *repeat_text_loc = NULL;
  int repeat_min;
  int repeat_max;
  int repeats;
  int tags_processed;
  int vals, nesting;
  char *repeat_tag_nest_loc = NULL;
  int repeat_text_len = 0;

  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  start:

  nesting = 0;
  tags_processed = 0;
  prep_buf_loc = prep_buf;
  send_buf_loc = send_buf;

  while ((repeat_tag_loc = strstr(send_buf_loc, repeat_tag)))
    {
    tags_processed = 1;

    /* check for nesting tags */
    while ((repeat_tag_nest_loc = strstr(repeat_tag_loc + strlen(repeat_tag),
        repeat_tag)))
      {
      repeat_tag_loc = repeat_tag_nest_loc;
      nesting = 1;
      }

    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, repeat_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += (repeat_tag_loc - send_buf_loc);

    /* advance send_buf_loc to end of tag */
    send_buf_loc = repeat_tag_loc + strlen(repeat_tag);

    /* parse repeat tag data */
    vals = sscanf(send_buf_loc, "%d %d", &repeat_min, &repeat_max);
    if (vals < 1)
      goto end;
    if (vals == 1 && repeat_min < 1)
      goto end;
    if (vals == 2 && (repeat_min < 1 || repeat_max < 1 || repeat_min
        > repeat_max))
      goto end;

    /* location of repeat text */
    if (!(repeat_text_loc = strstr(send_buf_loc, ">")))
      goto end;
    repeat_text_loc++;

    /* advance send_buf_loc to end of tag data */
    if (!(send_buf_loc = strstr(send_buf_loc, repeat_tag_end)))
      goto end;

    /* check that location of repeat text didn't fail */
    if (repeat_text_loc > send_buf_loc)
      goto end;

    /* calculate repeat text length */
    repeat_text_len = send_buf_loc - repeat_text_loc;

    /* set number of repetitions */
    if (vals == 2)
      repeats = rand() % ((1 + repeat_max - repeat_min)) + repeat_min;
    else
      repeats = repeat_min;

    /* advance send_buf_loc to end of end_tag */
    send_buf_loc += strlen(repeat_tag_end);

    /* resize prep_buf */
    prep_buf_loc_offset = prep_buf_loc - prep_buf;
    prep_buf_len = prep_buf_len + repeats * repeat_text_len - (send_buf_loc
        - repeat_tag_loc);
    prep_buf = tj_realloc(prep_buf, prep_buf_len);
    prep_buf_loc = prep_buf + prep_buf_loc_offset;

    /* start repetition */
    int i;
    for (i = 1; i <= repeats; i++)
      {
      memcpy(prep_buf_loc, repeat_text_loc, repeat_text_len);
      prep_buf_loc += repeat_text_len;
      }
    }

  if (tags_processed)
    {

    /* add last part from send buf to prep buf */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy prep buf to send buf */
    send_buf_len = prep_buf_len;
    send_buf = tj_realloc(send_buf, send_buf_len);
    memcpy(send_buf, prep_buf, send_buf_len);
    }

  if (nesting == 1)
    goto start;
  else
    end:

    free(prep_buf);
  *send_buf_p = send_buf;
  return send_buf_len;
  }

int instance_num_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len)
  {
  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *inst_tag = "<inst>";
  char *inst_tag_loc = NULL;
  int tags_processed = 0;
  char instance_str[8];

  while ((inst_tag_loc = strstr(send_buf_loc, inst_tag)))
    {
    tags_processed = 1;

    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, inst_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += (inst_tag_loc - send_buf_loc);

    /* advance send_buf_loc to end of tag */
    send_buf_loc = inst_tag_loc + strlen(inst_tag);

    /* insert instance number */
    sprintf(instance_str, "%d", instance_data->instance_num);
    memcpy(prep_buf_loc, instance_str, strlen(instance_str));
    prep_buf_loc += strlen(instance_str);

    /* resize prep_buf */
    prep_buf_loc_offset = prep_buf_loc - prep_buf;
    prep_buf_len = prep_buf_len - (send_buf_loc - inst_tag_loc) + 1;
    prep_buf = tj_realloc(prep_buf, prep_buf_len);
    prep_buf_loc = prep_buf + prep_buf_loc_offset;
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy prep buf to send buf */
    send_buf_len = prep_buf_len;
    send_buf = tj_realloc(send_buf, send_buf_len);
    memcpy(send_buf, prep_buf, send_buf_len);
    }
  free(prep_buf);
  *send_buf_p = send_buf;
  return send_buf_len;
  }

int char_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len)
  {
  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *char_tag = "<char ";
  char *char_tag_loc = NULL;
  int i, tags_processed = 0;
  int vals, char_hex, count;

  while ((char_tag_loc = strstr(send_buf_loc, char_tag)))
    {
    tags_processed = 1;

    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, char_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += (char_tag_loc - send_buf_loc);

    /* advance send_buf_loc to end of tag */
    send_buf_loc = char_tag_loc + strlen(char_tag);

    /* get tag data */
    vals = sscanf(send_buf_loc, "%x %d", &char_hex, &count);
    if (vals == 1 || count < 1)
      count = 1;

    /* advance send_buf_loc to end of tag */
    if (!(send_buf_loc = strstr(send_buf_loc, ">")))
      goto done;
    send_buf_loc++;

    /* resize prep_buf */
    prep_buf_loc_offset = prep_buf_loc - prep_buf;
    prep_buf_len = prep_buf_len + count - (send_buf_loc - char_tag_loc);
    prep_buf = tj_realloc(prep_buf, prep_buf_len);
    prep_buf_loc = prep_buf + prep_buf_loc_offset;

    /* insert characters */
    for (i = 1; i <= count; i++)
      {
      *prep_buf_loc = (unsigned char) char_hex;
      prep_buf_loc++;
      }
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy prep buf to send buf */
    send_buf_len = prep_buf_len;
    send_buf = tj_realloc(send_buf, send_buf_len);
    memcpy(send_buf, prep_buf, send_buf_len);
    }

  *send_buf_p = send_buf;
  done:

  free(prep_buf);
  return send_buf_len;
  }

int fuzz_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len)
  {
  char *prep_buf = NULL;
  prep_buf = tj_malloc(prep_buf, send_buf_len);
  char *prep_buf_loc = prep_buf;
  int prep_buf_loc_offset;
  int prep_buf_len = send_buf_len;

  char *send_buf = *send_buf_p;
  char *send_buf_loc = send_buf;
  char *fuzz_tag = "<fuzz ";
  char *fuzz_tag_loc = NULL;
  char tok_fuzz_type[16];
  int fuzz_buf_len;
  int tags_processed = 0;
  int i, vals;
  int fuzz_tag_min, fuzz_tag_max;
  int fuzz_mode = 0;

  while ((fuzz_tag_loc = strstr(send_buf_loc, fuzz_tag)))
    {
    tags_processed = 1;

    /* add text up to tag */
    memcpy(prep_buf_loc, send_buf_loc, fuzz_tag_loc - send_buf_loc);

    /* advance prep_buf_loc to end */
    prep_buf_loc += (fuzz_tag_loc - send_buf_loc);

    /* advance send_buf_loc to end of tag */
    send_buf_loc = fuzz_tag_loc + strlen(fuzz_tag);

    /* get tag data */
    vals = sscanf(send_buf_loc, "%s %d %d", tok_fuzz_type, &fuzz_tag_min,
        &fuzz_tag_max);
    if (vals < 2 || !(fuzz_mode = get_fuzz_type(tok_fuzz_type, instance_data)))
      goto done;
    if (vals == 2 && (fuzz_tag_min < 1))
      goto done;
    if (vals == 3 && (fuzz_tag_min < 1 || fuzz_tag_max < 1 || fuzz_tag_min
        > fuzz_tag_max))
      goto done;

    /* advance send_buf_loc to end of tag */
    if (!(send_buf_loc = strstr(send_buf_loc, ">")))
      goto done;
    send_buf_loc++;

    /* set fuzzing length */
    if (vals == 3)
      fuzz_buf_len = rand() % (1 + (fuzz_tag_max - fuzz_tag_min))
          + fuzz_tag_min - 1;
    else
      fuzz_buf_len = fuzz_tag_min;

    /* resize prep_buf */
    prep_buf_loc_offset = prep_buf_loc - prep_buf;
    prep_buf_len = prep_buf_len + fuzz_buf_len - (send_buf_loc - fuzz_tag_loc);
    prep_buf = tj_realloc(prep_buf, prep_buf_len);
    prep_buf_loc = prep_buf + prep_buf_loc_offset;

    /* set random fuzzing type */
    if (fuzz_mode == r_random)
      fuzz_mode = rand() % 9 + 1;

    /* insert characters */
    for (i = 1; i <= fuzz_buf_len; i++)
      {
      *prep_buf_loc = fuzz_char(fuzz_mode);
      prep_buf_loc++;
      }
    }

  if (tags_processed)
    {
    /* add last part */
    memcpy(prep_buf_loc, send_buf_loc, (send_buf + send_buf_len) - send_buf_loc);

    /* copy prep buf to send buf */
    send_buf_len = prep_buf_len;
    send_buf = tj_realloc(send_buf, send_buf_len);
    memcpy(send_buf, prep_buf, send_buf_len);
    }
  *send_buf_p = send_buf;
  done:

  free(prep_buf);
  return send_buf_len;
  }
