/*
 *  Tcpjunk - TCP protocols testing tool
 *  Contact: oferkv@gmail.com
 *
 *  Tcpjunk is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2, or (at your option) any
 *  later version.
 *
 *  Tcpjunk is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */

#include <openssl/crypto.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#define VER "2.660"
#define SOCK_BUF_LEN 16384
#define GUI_RDELAY "2"
#define GUI_SDELAY "100"
#define BACKLOG 10
#define MAX_SECTS 100
#define SERVER_CERT "server.crt"
#define SERVER_KEY "server.key"
#define SECT_START "<section>"
#define SECT_END "</section>"

enum fuzzmode
  {
  r_none,
  r_digit,
  r_alnum,
  r_alpha,
  r_cntrl,
  r_print,
  r_special,
  r_punct,
  r_space,
  r_any,
  r_random
  };

struct tj_data
  {
    char *dest_addr;
    int dest_addr_len;
    char *src_addr;
    int src_addr_len;
    int dest_port;
    int src_port;
    char interface[32];
    int offset;
    int rand_ip;
    char src_ip[32];
    char *session_data;
    int session_data_len;
    int debug;
    int wait;
    int itimeout;
    int stimeout;
    unsigned int keepalive;
    unsigned int loop;
    int new_sock;
    char *sections[MAX_SECTS];
    int sections_len[MAX_SECTS];
    int sect_idx;
    int wait_conn;
    int no_wait_last;
    int tcp_opt;
    int use_ssl;
    char server_cert[32];
    char server_key[32];
    int gui;
    char session_file_name[FILENAME_MAX];
    unsigned int instance_num;
    int segment_size;
    int segment_delay;
    int verbose;
    int rand_file;
    char **rand_file_list;
    int rand_files_num;
  };

/* junk.c */
void junk(struct tj_data children_args);

/* sjunk.c */
void sjunk(struct tj_data child_data);

/* utils.c */
int tj_connect(int sock, struct sockaddr_in sa, struct tj_data *children_args);
void tj_disconnect(int sock, struct tj_data *instance_data, SSL * ssl,
    SSL_CTX * ctx);
int read_reply(int sock, SSL *ssl, struct tj_data *children_args);
void send_data(int sock, SSL *ssl, struct tj_data *children_args,
    char *send_buf, int send_buf_len);
char fuzz_char(int ftype);
void sigchld_handler(int s);
int check_address(struct tj_data *instance_data, char *addr);
void load_session_data(struct tj_data *instance_data);
long tj_filesize(FILE *stream);
int check_port(int port, struct tj_data *instance_data);
void load_rand_file_list(struct tj_data *instance_data);
void set_random_session_file(struct tj_data *instance_data);
void find_sections(struct tj_data *instance_data);
void tj_error(struct tj_data *instance_data, char *message, char *param, int color);
void set_src_ip(struct tj_data *instance_data, char *bind_if_name,
    char *new_src_ip);
void show_help();
char*tj_malloc(char *buf, int buf_len);
char *tj_realloc(char *buf, int buf_len);
int get_fuzz_type(char *fuzz_type, struct tj_data *instance_data);
void cleanup(int sig);

/* gui.c */
void gui_init(int argc, char *argv[], struct tj_data instance_data);
void gui_show_traffic(char *buf, int buf_len, int color);
void gui_add_log(char *buf, int color);
void gui_accept();

/* tags.c */
int fuzz_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len);
int system_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len);
int repeat_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len);
int instance_num_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len);
int char_data(char **send_buf_p, struct tj_data *instance_data,
    int send_buf_len);

